;Tutorial 9.2 - Nigel Goodwin 2003
;Keypad reading with RS232 output

	LIST	p=16F628		;tell assembler what chip we are using
	include "P16F628.inc"		;include the defaults for the chip
	ERRORLEVEL	0,	-302	;suppress bank selection messages
	__config 0x3D18			;sets the configuration settings (oscillator type etc.)




		cblock	0x20			;start of general purpose registers
			count			;used in looping routines
			count1			;used in delay routine
			counta			;used in delay routine
			countb			;used in delay routine
			tmp1			;temporary storage
			tmp2

			key
			rows
			index
			Xmit_Byte        	;holds byte to xmit
            		Rcv_Byte        	;holds received byte 
            		Bit_Cntr        	;bit counter for RS232
            		Delay_Count        	;delay loop counter	
		endc

SER_PORT	Equ	PORTA
SER_TRIS	Equ	TRISA
SER_IN		Equ	0x07
SER_OUT		Equ	0x06
KEY_PORT	Equ	PORTB			;keypad port
KEY_TRIS	Equ	TRISB
Col1		Equ	0			;pins used for keypad inputs
Col2		Equ	1
Col3		Equ	2
Col4		Equ	3

		org	0x0000
		goto	Start

Key_Table  	ADDWF   PCL       , f
            	RETLW   0x43
            	RETLW   0x42
            	RETLW   0x30
            	RETLW   0x41
            	RETLW   0x44
            	RETLW   0x39
            	RETLW   0x38
            	RETLW   0x37
            	RETLW   0x45
            	RETLW   0x36
            	RETLW   0x35
            	RETLW   0x34
            	RETLW   0x46
            	RETLW   0x33
            	RETLW   0x32
            	RETLW   0x31


HEX_Table  	ADDWF   PCL       , f
            	RETLW   0x30
            	RETLW   0x31
            	RETLW   0x32
            	RETLW   0x33
            	RETLW   0x34
            	RETLW   0x35
            	RETLW   0x36
            	RETLW   0x37
            	RETLW   0x38
            	RETLW   0x39
            	RETLW   0x41
            	RETLW   0x42
            	RETLW   0x43
            	RETLW   0x44
            	RETLW   0x45
            	RETLW   0x46


Text		addwf	PCL, f
		retlw	'K'
		retlw	'e'
		retlw	'y'
		retlw	'p'
		retlw	'a'
		retlw	'd'
		retlw	' '
		retlw	'T'
		retlw	'e'
		retlw	's'
		retlw	't'
		retlw	'.'
		retlw	0x0A			;send LF
		retlw	0x0D			;send CR
		retlw	0x00

Start		movlw	0x07
		movwf	CMCON			;turn comparators off (make it like a 16F84)

Initialise	clrf	count
		clrf	PORTA
		clrf	PORTB




SetPorts	bsf 	STATUS,		RP0	;select bank 1
		movlw	0x00			;make all pins outputs
		movwf	SER_TRIS
		movlw	0x0F			;set keypad pins
		movwf	KEY_TRIS		;half in, half out
		bcf 	STATUS,		RP0	;select bank 0

		call	SER_INIT		;setup serial port
		call	Delay255		;a;;ow time to settle


		clrf	count			;set counter register to zero
Message		movf	count, w		;put counter value in W
		call	Text			;get a character from the text table
		xorlw	0x00			;is it a zero?
		btfsc	STATUS, Z
		goto	Main
		call	XMIT_RS232
		incf	count, f
		goto	Message

Main		movlw	0x0F
		movwf	KEY_PORT		;set all output pins low

Loop		call	Chk_Keys		;check keys
		movlw	'H'
		call	XMIT_RS232
		movlw	'e'
		call	XMIT_RS232
		movlw	'x'
		call	XMIT_RS232
		movlw	' '
		call	XMIT_RS232
		movf	key, w			;restore key pressed
		call	XMIT_HEX		;display in hex
		movlw	' '
		call	XMIT_RS232
		movlw	'A'
		call	XMIT_RS232
		movlw	'S'
		call	XMIT_RS232
		movlw	'C'
		call	XMIT_RS232
		movlw	'I'
		call	XMIT_RS232
		movlw	'I'
		call	XMIT_RS232
		movlw	' '
		call	XMIT_RS232
		movf	key, w			;restore key pressed
		call	XMIT_RS232		;display as ASCII
		movlw	0x0A			;LF
		call	XMIT_RS232
		movlw	0x0D			;CR
		call	XMIT_RS232
		goto	Loop

;Keypad subroutine

Chk_Keys	movlw	0x00			;wait until no key pressed
		movwf	KEY_PORT		;set all output pins low
		movf	KEY_PORT,	W
		andlw	0x0F			;mask off high byte
		sublw	0x0F
		btfsc	STATUS, Z		;test if any key pressed
		goto	Keys			;if none, read keys
		call	Delay20
		goto	Chk_Keys		;else try again

Keys	  	call    Scan_Keys
            	movlw   0x10			;check for no key pressed
            	subwf   key, w
            	btfss   STATUS, Z
            	goto    Key_Found
		call	Delay20
		goto	Keys
Key_Found      	movf    key, w
		andlw	0x0f
		call	Key_Table		;lookup key in table	
		movwf	key			;save back in key
		return				;key pressed now in W

Scan_Keys  	clrf    key
		movlw	0xF0			;set all output lines high
            	movwf   KEY_PORT
            	movlw   0x04
            	movwf   rows			;set number of rows
            	bcf     STATUS, C		;put a 0 into carry
Scan	  	rrf     KEY_PORT, f
            	bsf     STATUS, C		;follow the zero with ones
;comment out next two lines for 4x3 numeric keypad.
            	btfss   KEY_PORT, Col4
            	goto    Press
            	incf    key, f
            	btfss   KEY_PORT, Col3
            	goto    Press
            	incf    key, f
            	btfss   KEY_PORT, Col2
            	goto    Press
            	incf    key, f
            	btfss   KEY_PORT, Col1
            	goto    Press
            	incf    key, f
            	decfsz  rows, f
            	goto    Scan
Press	  	return

;end of keypad subroutines.


;Serial routines

SER_INIT
            	BSF     STATUS, RP0           ;select bank 1
            	BCF     SER_TRIS, SER_OUT     ;set B6 as an output
            	BSF     SER_TRIS, SER_IN      ;set B7 as an input
            	BCF     STATUS, RP0           ;select bank 0
            	BSF     SER_PORT, SER_OUT     ;set SER_OUT high
            	RETURN

XMIT_RS232  	MOVWF   Xmit_Byte             ;move W to Xmit_Byte
            	MOVLW   0x08                  ;set 8 bits out
            	MOVWF   Bit_Cntr
            	BCF     SER_PORT, SER_OUT
            	CALL    Bit_Delay
Ser_Loop    	RRF     Xmit_Byte , f         ;send one bit
            	BTFSS   STATUS    , C
            	BCF     SER_PORT, SER_OUT
            	BTFSC   STATUS    , C
            	BSF     SER_PORT, SER_OUT
            	CALL    Bit_Delay
            	DECFSZ  Bit_Cntr  , f         ;test if all done
            	GOTO    Ser_Loop
            	BSF     SER_PORT, SER_OUT
            	CALL    Bit_Delay
            	RETURN

Rcv_RS232   	BTFSC   SER_PORT, SER_IN      ;wait for start bit
            	GOTO    Rcv_RS232
            	CALL    Start_Delay	          ;do half bit time delay
            	BTFSC   SER_PORT, SER_IN      ;check still in start bit
            	GOTO    Rcv_RS232
            	MOVLW   0x08                  ;set up to read 8 bits
            	MOVWF   Bit_Cntr
            	CLRF    Rcv_Byte
Next_RcvBit 	CALL    Bit_Delay
            	BTFSS   SER_PORT, SER_IN
            	BCF     STATUS    , C
            	BTFSC   SER_PORT, SER_IN
            	BSF     STATUS    , C
            	RRF     Rcv_Byte  , f
            	DECFSZ  Bit_Cntr  , f         ;test if all done
            	GOTO    Next_RcvBit
            	CALL    Bit_Delay
            	MOVF    Rcv_Byte, W
            	RETURN

Start_Delay 	MOVLW   0x0C
            	MOVWF   Delay_Count
Start_Wait  	NOP
            	DECFSZ  Delay_Count , f
            	GOTO    Start_Wait
            	RETURN

Bit_Delay   	MOVLW   0x18
            	MOVWF   Delay_Count
Bit_Wait    	NOP
            	DECFSZ  Delay_Count , f
            	GOTO    Bit_Wait
            	RETURN

XMIT_HEX	movwf	tmp1
		swapf	tmp1,	w
		andlw	0x0f
		call	HEX_Table
		call	XMIT_RS232
		movf	tmp1, w
		andlw	0x0f
		call	HEX_Table
		call	XMIT_RS232
		retlw	0x00

;End of serial routines


Delay255	movlw	0xff			;delay 255 mS
		goto	d0
Delay100	movlw	d'100'			;delay 100mS
		goto	d0
Delay50		movlw	d'50'			;delay 50mS
		goto	d0
Delay20		movlw	d'20'			;delay 20mS
		goto	d0
Delay5		movlw	0x05			;delay 5.000 ms (4 MHz clock)
d0		movwf	count1
d1		movlw	0xC7			;delay 1mS
		movwf	counta
		movlw	0x01
		movwf	countb
Delay_0
		decfsz	counta, f
		goto	$+2
		decfsz	countb, f
		goto	Delay_0

		decfsz	count1	,f
		goto	d1
		retlw	0x00


		end